'''Sample k-space trajectories.'''

from functools import partial

import numpy as np

def radial(sx, spokes, golden=False):
    '''k-space coordinates for a radial trajectory.

    Parameters
    ----------
    sx : int
        Number of samples along spoke.
    spokes : int
        Number of spokes.
    golden : bool, optional
        Use Golden angle ratio sampling.

    Returns
    -------
    kx, ky : array_like
        k-space coordinates for sampling along a radial trajectory.

    Notes
    -----
    This is a very simple radial trajectory mainly for demonstration
    purposes.
    '''

    N = sx*spokes
    kx = np.zeros(N)
    ky = np.zeros(N)
    x = np.linspace(-1, 1, sx)*sx/2 # scale fac to match BART's traj
    y = np.zeros(sx)

    # How we get
    if golden:
        GA = np.pi*(3 - np.sqrt(5)) # calculate GA for use if needed
        gettheta = partial(_nextspoke_golden, GA=GA)
    else:
        gettheta = partial(_nextspoke, spokes=spokes)

    for ii in range(spokes):
        # theta = ii/spokes*np.pi
        theta = gettheta(ii)
        ct = np.cos(theta)
        st = np.sin(theta)
        kx[ii*sx:(ii+1)*sx] = x*ct - y*st
        ky[ii*sx:(ii+1)*sx] = x*st + y*ct
    return(kx, ky)

def _nextspoke(ii, spokes):
    return ii/spokes*np.pi

def _nextspoke_golden(ii, GA):
    return GA*ii



'''Simple dynamic numerical phantom.'''

import numpy as np

def dynamic(N, nt):
    '''Concentric circles with dynamic movement.

    Parameters
    ----------
    N : int
        In-plane dimension.
    nt : int
        Number of time points.

    Returns
    -------
    ph : array_like
        Phantom of size (N, N, nt), time is last dimension.
    '''

    ph = np.zeros((N, N, nt))
    X, Y = np.meshgrid(
        np.linspace(-1, 1, N),
        np.linspace(-1, 1, N))

    # Make an outer circle
    thickness = .25
    idx0 = X**2 + Y**2 <= 1
    idx1 = X**2 + Y**2 >= (1 - thickness)**2
    idx = np.logical_and(idx0, idx1)
    ph[idx, :] = 1

    # Make another outer circle
    idx0 = X**2 + Y**2 <= (1 - thickness)**2
    idx1 = X**2 + Y**2 >= (1 - 2*thickness)**2
    idx = np.logical_and(idx0, idx1)
    ph[idx, :] = .2

    # Make inner circle
    r = np.cos(np.arange(nt)*2*np.pi/nt) + 2
    r = r/np.max(r)*.4
    thickness = .15
    for ii in range(nt):
        idx0 = X**2 + Y**2 <= r[ii]**2
        idx1 = X**2 + Y**2 >= (r[ii] - thickness)**2
        idx = np.logical_and(idx0, idx1)
        ph[idx, ii] = .8

    return ph


if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-
'''The canonical Shepp-Logan phantom used for CT simulations.'''

import numpy as np

def ct_shepp_logan(
        N, modified=True, E=None, ret_E=False, zlims=(-1, 1)):
    '''Generate a Shepp-Logan phantom of size (N, N).

    Parameters
    ----------
    N : int or array_like
        Matrix size, (N, N) or (M, N) or (L, M, N).
    modified : bool, optional
        Use original grey-scale values as given in [1]_.  Most
        implementations use modified values for better contrast (for
        example, see [3]_ and [4]_).
    E : array_like, optional
        For 2D: ex6 numeric matrix defining e ellipses.  The six
        columns of E are:

            - Length of the horizontal semiaxis of the ellipse
            - Length of the vertical semiaxis of the ellipse
            - x-coordinate of the center of the ellipse (in [-1, 1])
            - y-coordinate of the center of the ellipse (in [-1, 1])
            - Angle between the horizontal semiaxis of the ellipse
              and the x-axis of the image (in rad)

        For 3D: ex8 numeric matrix defining e ellipsoids.  The
        columns are:

            - Additive intensity value of the ellipse
            - x principal axis of the ellipsoid
            - y principal axis of the ellipsoid
            - z principal axis of the ellipsoid
            - x-coordinate of the center of the ellipsoid (in [-1, 1])
            - y-coordinate of the center of the ellipsoid (in [-1, 1])
            - z-coordinate of the center of the ellipsoid (in [-1, 1])
            - Angle of the ellipsoid (in rad)

    ret_E : bool, optional
        Return the matrix E used to generate the phantom, ph.
    zlims : tuple, optional
        Only for 3D.  Specify bounds along z.  Often we only want the
        middle portion of a 3D phantom, e.g., zlim=(-.5, .5).

    Returns
    -------
    ph : array_like
        The Shepp-Logan phantom.
    E : array_like, optional
        The ellipse parameters used to generate ph.

    Notes
    -----
    This much abused phantom is due to [1]_.  The tabulated values in
    the paper are reproduced in the Wikipedia entry [2]_.  The
    original values do not produce great contrast, so modified values
    are used by default (see Table B.1 in [5]_ or implementations
    [3]_ and [4]_).

    The 3D version follows values tabulated in [6]_.  The modified
    version uses the same adjustments to grayscale.

    References
    ----------
    .. [1] Shepp, Lawrence A., and Benjamin F. Logan. "The Fourier
           reconstruction of a head section." IEEE Transactions on
           nuclear science 21.3 (1974): 21-43.
    .. [2] https://en.wikipedia.org/wiki/Shepp%E2%80%93Logan_phantom
    .. [3] https://sigpy.readthedocs.io/en/latest/_modules/sigpy/
           sim.html#shepp_logan
    .. [4] http://www.mathworks.com/matlabcentral/fileexchange/
           9416-3d-shepp-logan-phantom
    .. [5] Toft, Peter Aundal, and John Aasted Sørensen. "The Radon
           transform-theory and implementation." (1996).
    .. [6] Koay, Cheng Guan, Joelle E. Sarlls, and Evren Özarslan.
           "Three‐dimensional analytical magnetic resonance imaging
           phantom in the Fourier domain." Magnetic Resonance in
           Medicine: An Official Journal of the International Society
           for Magnetic Resonance in Medicine 58.2 (2007): 430-436.
    '''

    # Get size of phantom
    if np.isscalar(N):
        M, N = N, N
        is2D = True
    else:
        if len(N) == 2:
            M, N = N[:]
            is2D = True
        elif len(N) == 3:
            L, M, N = N[:]
            is2D = False
        else:
            raise ValueError('Dimension must be scalar, 2D, or 3D!')

    # Give back either a 2D or 3D phantom
    if is2D:
        return ct_shepp_logan_2d(M, N, modified, E, ret_E)
    return ct_shepp_logan_3d(L, M, N, modified, E, ret_E, zlims)

def ct_shepp_logan_2d(M, N, modified, E, ret_E):
    '''Make a 2D phantom.'''

    # Get the ellipse parameters the user asked for
    if E is None:
        if modified:
            E = _modified_shepp_logan_params_2d()
        else:
            E = _shepp_logan_params_2d()

    # Extract params
    grey = E[:, 0]
    major = E[:, 1]
    minor = E[:, 2]
    xs = E[:, 3]
    ys = E[:, 4]
    theta = E[:, 5]

    # 2x2 square => FOV = (-1, 1)
    X, Y = np.meshgrid( # meshgrid needs linspace in opposite order
        np.linspace(-1, 1, N),
        np.linspace(-1, 1, M))
    ph = np.zeros((M, N))
    ct = np.cos(theta)
    st = np.sin(theta)

    for ii in range(E.shape[0]):
        xc, yc = xs[ii], ys[ii]
        a, b = major[ii], minor[ii]
        ct0, st0 = ct[ii], st[ii]

        # Find indices falling inside the ellipse
        idx = (
            ((X - xc)*ct0 + (Y - yc)*st0)**2/a**2 +
            ((X - xc)*st0 - (Y - yc)*ct0)**2/b**2 <= 1)

        # Sum of ellipses
        ph[idx] += grey[ii]

    if ret_E:
        return(ph, E)
    return ph

def ct_shepp_logan_3d(L, M, N, modified, E, ret_E, zlims):
    '''Make a 3D phantom.'''

    # Make sure zlims are appropriate
    assert len(zlims) == 2, (
        'zlims must be a tuple with 2 entries: upper and lower '
        'bounds!')
    assert zlims[0] <= zlims[1], (
        'zlims: lower bound must be first entry!')

    # Get parameters from paper if None provided
    if E is None:
        if modified:
            E = _modified_shepp_logan_params_3d()
        else:
            E = _shepp_logan_params_3d()

    # Extract some parameters so we can use them
    gray = E[:, 0]
    xaxis = E[:, 1]
    yaxis = E[:, 2]
    zaxis = E[:, 3]
    xs = E[:, 4]
    ys = E[:, 5]
    zs = E[:, 6]
    theta = E[:, 7]

    # Initialize array
    X, Y, Z = np.meshgrid( # meshgrid does X, Y backwards
        np.linspace(-1, 1, M),
        np.linspace(-1, 1, L),
        np.linspace(zlims[0], zlims[1], N))
    ct = np.cos(theta)
    st = np.sin(theta)
    ph = np.zeros((L, M, N))

    # Put ellipses where they need to be
    for ii in range(E.shape[0]):
        xc, yc, zc = xs[ii], ys[ii], zs[ii]
        a, b, c = xaxis[ii], yaxis[ii], zaxis[ii]
        ct0, st0 = ct[ii], st[ii]

        # Find indices falling inside the ellipsoid, ellipses only
        # rotated in xy plane
        idx = (
            ((X - xc)*ct0 + (Y - yc)*st0)**2/a**2 +
            ((X - xc)*st0 - (Y - yc)*ct0)**2/b**2 +
            (Z - zc)**2/c**2 <= 1)

        # Add ellipses together
        ph[idx] += gray[ii]

    if ret_E:
        return(ph, E)
    return ph

def _shepp_logan_params_2d():
    '''Return parameters for original Shepp-Logan phantom.

    Returns
    -------
    E : array_like, shape (10, 6)
        Parameters for the 10 ellipses used to construct the phantom.
    '''

    E = np.zeros((10, 6)) # (10, [A, a, b, xc, yc, theta])
    E[:, 0] = [2, -.98, -.02, -.02, .01, .01, .01, .01, .01, .01]
    E[:, 1] = [
        .69, .6624, .11, .16, .21, .046, .046, .046, .023, .023]
    E[:, 2] = [.92, .874, .31, .41, .25, .046, .046, .023, .023, .046]
    E[:, 3] = [0, 0, .22, -.22, 0, 0, 0, -.08, 0, .06]
    E[:, 4] = [0, -.0184, 0, 0, .35, .1, -.1, -.605, -.605, -.605]
    E[:, 5] = np.deg2rad([0, 0, -18, 18, 0, 0, 0, 0, 0, 0])
    return E

def _modified_shepp_logan_params_2d():
    '''Return parameters for modified Shepp-Logan phantom.

    Returns
    -------
    E : array_like, shape (10, 6)
        Parameters for the 10 ellipses used to construct the phantom.
    '''
    E = _shepp_logan_params_2d()
    E[:, 0] = [1, -0.8, -0.2, -0.2, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1]
    return E

def _shepp_logan_params_3d():
    '''Ellipsoid parameters for 3D Shepp-Logan.

    Returns
    -------
    E : array_like, shape (10, 8)
        Parameters for the 10 ellipsoids used to construct phantom.
    '''

    E = np.zeros((10, 8))
    # [ gray, x, y, z, a, b, c, theta ]
    E[0, :] = [2, 0.69, 0.92, 0.9, 0, 0, 0, 0]
    E[1, :] = [-.8, 0.6624, 0.874, 0.88, 0, 0, 0, 0]
    E[2, :] = [-.2, 0.41, 0.16, 0.21, -0.22, 0, -0.25, 3*np.pi/5]
    E[3, :] = [-.2, 0.31, 0.11, 0.22, 0.22, 0, -0.25, 2*np.pi/5]
    E[4, :] = [.2, 0.21, 0.25, 0.5, 0, 0.35, -0.25, 0]
    E[5, :] = [.2, 0.046, 0.046, 0.046, 0, 0.1, -0.25, 0]
    E[6, :] = [.1, 0.046, 0.023, 0.02, -0.08, -0.65, -0.25, 0]
    E[7, :] = [.1, 0.046, 0.023, 0.02, 0.06, -0.65, -0.25, np.pi/2]
    E[8, :] = [.2, 0.056, 0.04, 0.1, 0.06, -0.105, 0.625, np.pi/2]
    E[9, :] = [-.2, 0.056, 0.056, 0.1, 0, 0.1, 0.625, 0]
    return E

def _modified_shepp_logan_params_3d():
    '''Return parameters for modified Shepp-Logan phantom.

    Returns
    -------
    E : array_like, shape (10, 8)
        Parameters for the 10 ellipsoids used to construct phantom.
    '''
    E = _shepp_logan_params_3d()
    E[:, 0] = [1, -0.8, -0.2, -0.2, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1]
    return E


if __name__ == '__main__':
    pass



'''Bring functions up to the correct level.'''

from .ct_shepp_logan import ct_shepp_logan
from .mr_shepp_logan import mr_shepp_logan
from .shepp_logan import shepp_logan
from .dynamic import dynamic
from .kspace import kspace_shepp_logan
from .traj import radial



# -*- coding: utf-8 -*-
'''Shepp-Logan phantom for use with MR simulations.'''

import numpy as np

def mr_shepp_logan(N, E=None, B0=3, T2star=False, zlims=(-1, 1)):
    '''Shepp-Logan phantom with MR tissue parameters.

    Parameters
    ----------
    N : int or array_like
        Matrix size, (N, N, N), or (L, M, N).
    E : array_like, optional
        ex13 numeric matrix defining e ellipses.  The columns of E
        are:

            - x-coordinate of the center of the ellipsoid (in [-1, 1])
            - y-coordinate of the center of the ellipsoid (in [-1, 1])
            - z-coordinate of the center of the ellipsoid (in [-1, 1])
            - x principal axis of the ellipsoid
            - y principal axis of the ellipsoid
            - z principal axis of the ellipsoid
            - Angle of the ellipsoid (in rad)
            - spin density, M0
            - Parameter A for T1 determination
            - Parameter C for T1 determination
            - Explicit T1 value (in sec, or np.nan if model is used)
            - T2 value (in sec)
            - chi (change in magnetic susceptibility)

        If spin density is negative, M0, T1, and T2 will be subtracted
        instead of cummulatively added.
    B0 : float, optimal
        Field strength (in Tesla).
    T2star : bool, optional
        Use magnetic susceptibility values to return T2star values
        instead of T2. Gyromagnetic ratio is assumed to be that of
        hydrogen.
    zlims : tuple, optional
        Only for 3D.  Specify bounds along z.  Often we only want the
        middle portion of a 3D phantom, e.g., zlim=(-.5, .5).

    Returns
    -------
    M0 : array_like
        The proton density.
    T1 : array_like
        The T1 values.
    T2 : array_like
        The T2 values. If T2star is True, then these will be T2 star
        values.

    Notes
    -----
    Implements the phantoms described in [1]_.

    If parameters A, C are given and T1 is None, T1 is determined
    according to the equation:

        T1 = A*B0^C

    References
    ----------
    .. [1] Gach, H. Michael, Costin Tanase, and Fernando Boada.
           "2D & 3D Shepp-Logan phantom standards for MRI." 2008 19th
           International Conference on Systems Engineering. IEEE,
           2008.
    '''

    # Determine size of phantom
    if np.isscalar(N):
        L, M, N = N, N, N
    else:
        L, M, N = N[:]

    # Make sure zlims are appropriate
    assert len(zlims) == 2, (
        'zlims must be a tuple with 2 entries: upper and lower '
        'bounds!')
    assert zlims[0] <= zlims[1], (
        'zlims: lower bound must be first entry!')

    # Get parameters from paper if None provided
    if E is None:
        E = _ellipsoid_parameters()

    # Extract some parameters so we can use them
    xs = E[:, 0]
    ys = E[:, 1]
    zs = E[:, 2]
    xaxis = E[:, 3]
    yaxis = E[:, 4]
    zaxis = E[:, 5]
    theta = E[:, 6]
    M0 = E[:, 7]
    As = E[:, 8]
    Cs = E[:, 9]
    T1 = E[:, 10]
    T2 = E[:, 11]
    chis = E[:, 12]

    # Initialize array
    X, Y, Z = np.meshgrid( # meshgrid does X, Y backwards
        np.linspace(-1, 1, M),
        np.linspace(-1, 1, L),
        np.linspace(zlims[0], zlims[1], N))
    ct = np.cos(theta)
    st = np.sin(theta)
    sgn = np.sign(M0)
    T1s = np.zeros((L, M, N))
    T2s = np.zeros((L, M, N))
    M0s = np.zeros((L, M, N))

    # We'll need the gyromagnetic ratio if returning T2star values
    if T2star:
        # see https://en.wikipedia.org/wiki/Gyromagnetic_ratio:
        gamma0 = 267.52219 # 10^6 rad⋅s−1⋅T⋅−1

    # Put ellipses where they need to be
    for ii in range(E.shape[0]):
        xc, yc, zc = xs[ii], ys[ii], zs[ii]
        a, b, c = xaxis[ii], yaxis[ii], zaxis[ii]
        ct0, st0 = ct[ii], st[ii]

        # Find indices falling inside the ellipsoid, ellipses only
        # rotated in xy plane
        idx = (
            ((X - xc)*ct0 + (Y - yc)*st0)**2/a**2 +
            ((X - xc)*st0 - (Y - yc)*ct0)**2/b**2 +
            (Z - zc)**2/c**2 <= 1)

        # Add ellipses together -- subtract of M0 is negative
        M0s[idx] += M0[ii]

        # Use T2star values if user asked for them
        if T2star:
            T2s[idx] += sgn[ii]/(1/T2[ii] + gamma0*np.abs(
                B0*chis[ii]))
        else:
            T2s[idx] += sgn[ii]*T2[ii]

        # Use T1 model if not given explicit T1 value
        if np.isnan(T1[ii]):
            T1s[idx] += sgn[ii]*As[ii]*(B0**Cs[ii])
        else:
            T1s[idx] += sgn[ii]*T1[ii]

    return(M0s, T1s, T2s)


def _ellipsoid_parameters():
    '''Returns parameters of ellipsoids.

    Returns
    -------
    E : array_like
        Parameters for the ellipsoids used to construct the phantom.
    '''
    params = _mr_relaxation_parameters()

    E = np.zeros((15, 13))
    # [:, [x, y, z, a, b, c, theta, m0, A, C, (t1), t2, chi]]
    E[0, :] = [
        0, 0, 0, .72, .95, .93, 0, .8,
        *params['scalp']]
    E[1, :] = [
        0, 0, 0, .69, .92, .9, 0, .12,
        *params['marrow']]
    E[2, :] = [
        0, -.0184, 0, .6624, .874, .88, 0, .98,
        *params['csf']]
    E[3, :] = [
        0, -.0184, 0, .6524, .864, .87, 0, .745,
        *params['gray-matter']]
    E[4, :] = [
        -.22, 0, -.25, .41, .16, .21, np.deg2rad(-72), .98,
        *params['csf']]
    E[5, :] = [
        .22, 0, -.25, .31, .11, .22, np.deg2rad(72), .98,
        *params['csf']]
    E[6, :] = [
        0, .35, -.25, .21, .25, .35, 0, .617,
        *params['white-matter']]
    E[7, :] = [
        0, .1, -.25, .046, .046, .046, 0, .95,
        *params['tumor']]
    E[8, :] = [
        -.08, -.605, -.25, .046, .023, .02, 0, .95,
        *params['tumor']]
    E[9, :] = [
        .06, -.605, -.25, .046, .023, .02, np.deg2rad(-90), .95,
        *params['tumor']]
    E[10, :] = [
        0, -.1, -.25, .046, .046, .046, 0, .95,
        *params['tumor']]
    E[11, :] = [
        0, -.605, -.25, .023, .023, .023, 0, .95,
        *params['tumor']]
    E[12, :] = [
        .06, -.105, .0625, .056, .04, .1, np.deg2rad(-90), .93,
        *params['tumor']]
    E[13, :] = [
        0, .1, .625, .056, .056, .1, 0, .98,
        *params['csf']]
    E[14, :] = [
        .56, -.4, -.25, .2, .03, .1, np.deg2rad(70), .85,
        *params['blood-clot']]

    # Need to subtract some ellipses here...
    Eneg = np.zeros(E.shape)
    for ii in range(E.shape[0]):

        # Ellipsoid geometry
        Eneg[ii, :7] = E[ii, :7]

        # Tissue property differs after 4th subtracted ellipsoid
        if ii > 3:
            Eneg[ii, 7:] = E[3, 7:]
        else:
            Eneg[ii, 7:] = E[ii-1, 7:]

    # Throw out first as we skip this one in the paper's table
    Eneg = Eneg[1:, :]

    # Spin density is negative for subtraction
    Eneg[:, 7] *= -1

    # Paper doesn't use last blood-clot ellipsoid
    E = E[:-1, :]
    Eneg = Eneg[:-1, :]

    # Put both ellipsoid groups together
    E = np.concatenate((E, Eneg), axis=0)

    return E

def _mr_relaxation_parameters():
    '''Returns MR relaxation parameters for certain tissues.

    Returns
    -------
    params : dict
        Gives entries as [A, C, (t1), t2, chi]

    Notes
    -----
    If t1 is None, the model T1 = A*B0^C will be used.  If t1 is not
    np.nan, then specified t1 will be used.
    '''

    # params['tissue-name'] = [A, C, (t1 value if explicit), t2, chi]
    params = dict()
    params['scalp'] = [.324, .137, np.nan, .07, -7.5e-6]
    params['marrow'] = [.533, .088, np.nan, .05, -8.85e-6]
    params['csf'] = [np.nan, np.nan, 4.2, 1.99, -9e-6]
    params['blood-clot'] = [1.35, .34, np.nan, .2, -9e-6]
    params['gray-matter'] = [.857, .376, np.nan, .1, -9e-6]
    params['white-matter'] = [.583, .382, np.nan, .08, -9e-6]
    params['tumor'] = [.926, .217, np.nan, .1, -9e-6]
    return params
